# -*- coding: utf-8 -*-
"""
/***************************************************************************
 CtrlCraft
                                 A QGIS plugin
 Areva - plugin for displaying 3D mines
                              -------------------
        begin                : 2017-09-12
        git sha              : $Format:%H$
        copyright            : (C) 2017 by Oslandia
        email                : sylvain.beorchia@oslandia.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""
from __future__ import absolute_import
from .settings import SIMU
from .simulator import CtrlCraftSimulator
from .project import CtrlCraftProject
from .lithology import CtrlCraftLithology
from .truck import CtrlCraftTruck, CtrlCraftTruckLoadMapTool
from .logger import CtrlCraftLogger
from .rtkube_proxy import RTKubeProxy
from .data import CtrlCraftData
from .widgets.level_filter import CtrlCraftLevelFilter
from .widgets.export_atlas import ExportAtlas
from .widgets.affect_radiometry_slab import CtrlCraftAffectRadiometrySlab
from .widgets.main_parameters_dock import CtrlCraftMainParametersDock
from .widgets.probe_radiometry_dock import CtrlCraftProbeRadiometryDock
from .widgets.trucks_loads_status_export import CtrlCraftTrucksLoadsStatusExport
from .widgets.slab_threshold import CtrlCraftSlabThreshold
from .widgets.database_import import CtrlCraftDatabaseImport
from .widgets.database_export import CtrlCraftDatabaseExport
from .widgets.viewer_3d.viewer_3d import CtrlCraftViewer3D
from .widgets.office_co3_import import CtrlCraftOfficeCO3Import
from .widgets.office_gate_import import CtrlCraftOfficeGateImport
from .widgets.office_slabs_import import CtrlCraftOfficeSlabsImport
from .widgets.office_estimated_blocs_chart import CtrlCraftOfficeEstimatedBlocsChart
from .widgets.office_estimated_blocs_import import CtrlCraftOfficeEstimatedBlocsImport
from .widgets.office_estimated_blocs_table import CtrlCraftOfficeEstimatedBlocsTable
from .widgets.office_shipments_chart import CtrlCraftOfficeShipmentsChart
from .widgets.office_shipments_table import CtrlCraftOfficeShipmentsTable
from . import resources
from builtins import object
import time
import os.path
import re
import configparser
from . import settings
import platform
from functools import partial

from PyQt5.QtCore import (QSettings,
                          QTranslator,
                          qVersion,
                          QCoreApplication,
                          Qt,
                          QSize)
from qgis.PyQt.QtWidgets import QApplication
from PyQt5.QtWidgets import (QComboBox,
                             QSizePolicy,
                             qApp,
                             QMenu,
                             QProgressBar,
                             QLabel,
                             QAction,
                             QWidget,
                             QMessageBox,
                             QDockWidget,
                             QHBoxLayout,
                             QDialog,
                             QPushButton)

from PyQt5.QtGui import (QPixmap,
                         QIcon)
from qgis.core import (QgsProject,
                       QgsMapLayerStyle,
                       QgsMessageLog)

if platform.system() == 'Windows':
    from qgis.utils import active_plugins

import re
import unicodedata


def strip_accents(text):
    """
    Strip accents from input String.

    :param text: The input string.
    :type text: String.

    :returns: The processed String.
    :rtype: String.
    """
    try:
        text = unicode(text, 'utf-8')
    except (TypeError, NameError):  # unicode is a default on python 3
        pass
    text = unicodedata.normalize('NFD', text)
    text = text.encode('ascii', 'ignore')
    text = text.decode("utf-8")
    return str(text)


def text_to_id(text):
    """
    Convert input text to id.

    :param text: The input string.
    :type text: String.

    :returns: The processed String.
    :rtype: String.
    """
    text = strip_accents(text.lower())
    text = re.sub('[ ]+', '_', text)
    text = re.sub('[^0-9a-zA-Z_-]', '', text)
    return text


# Initialize Qt resources from file resources.py

# Import the code for the dialog


class CtrlCraft(object):
    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        """Constructor.

        :param iface: An interface instance that will be passed to this class
            which provides the hook by which you can manipulate the QGIS
            application at run time.
        :type iface: QgsInterface
        """
        self.name = 'CtrlCraft'

        # Init logger
        self.logger = CtrlCraftLogger()
        self.logger.name = self.name
        self.logger.iface = iface

        version = self.tr('office')
        if CtrlCraft.isMobileVersion():
            version = self.tr('mobile')
        msg = self.tr('Initialize CtrlCraft plugin ')+'{}'.format(version)
        self.logger.log(msg)

        # Save reference to the QGIS interface
        self.viewMode = 2  # 2D only by default
        self.timer = None
        self.iface = iface
        self.canvas = self.iface.mapCanvas()
        self.data = None
        self.rtkube = None
        self.action_open = False
        self.truck = None

        # keep original view size for restoring
        self.viewSize = None
        self.viewMinSize = None
        self.viewMaxSize = None

        # Custom projection
        self.projIniFile = os.path.join(os.path.dirname(__file__), 'proj.ini')

        # Custom gamma correction
        self.gammaIniFile = os.path.join(
            os.path.dirname(__file__), 'gamma.ini')

        # initialize plugin directory
        plugin_dir = os.path.dirname(__file__)

        # initialize locale
        locale = QSettings().value('locale/userLocale')[0:2]
        self.iconSize = QSettings().value('IconSize')
        if not self.iconSize:
            self.iconSize = 24
        locale_path = os.path.join(
            plugin_dir,
            'widgets',
            'i18n',
            'CtrlCraft_{}.qm'.format(locale))

        if os.path.exists(locale_path):
            self.translator = QTranslator()
            self.translator.load(locale_path)

            if qVersion() > '4.3.3':
                QCoreApplication.installTranslator(self.translator)

        # init QSettings parameters
        self.initQSettingsParameters()
        # Declare instance attributes
        self.actions = []
        self.toolbar = self.iface.addToolBar(self.name)
        self.toolbar.setObjectName(self.name)

        self.toolbar_office = self.iface.addToolBar(
            '{}'.format(self.name) + self.tr(' Office'))
        self.toolbar_office.setObjectName(
            '{}'.format(self.name) + self.tr(' Office'))

        # Map canvas refresh management
        iface.mapCanvas().mapCanvasRefreshed.connect(self.canvasRefreshed)
        self.__waitingForRefresh = False

        # Check which project is loaded to activate some plugin
        # functionnalities accordingly
        QgsProject.instance().readProject.connect(self.projectLoaded)
        iface.newProjectCreated.connect(self.projectClosed)

        # Init data and try to connect to the underlying database
        self.data = CtrlCraftData()
        self.data.open()
        if not self.data.is_valid():
            self.data.create()

        # Create underlying project if necessary
        self.project = CtrlCraftProject()
        if not self.project.exists():
            self.project.create()

    @property
    def currentDeposit(self):
        return self.depositCombo.itemData(self.depositCombo.currentIndex())

    @property
    def currentPanel(self):
        return self.panelCombo.itemData(self.panelCombo.currentIndex())

    @property
    def currentBlast(self):
        return self.blastCombo.itemData(self.blastCombo.currentIndex())

    def refresh(self):
        CtrlCraftLogger().log()

        layer = self.project.slabsLayer()
        if layer:
            layer.triggerRepaint()
            self.iface.mapCanvas().refresh()

        self.viewer3d.refreshData(False)

    def truckUpdated(self):
        empty = True
        for truck in list(self.trucks.values()):
            empty &= truck.isEmpty()

        self.undo_action.setEnabled(not empty)

    # noinspection PyMethodMayBeStatic
    def tr(self, message):
        """Get the translation for a string using Qt translation API.

        We implement this ourselves since we do not inherit QObject.

        :param message: String for translation.
        :type message: str, QString

        :returns: Translated version of message.
        :rtype: QString
        """
        # noinspection PyTypeChecker,PyArgumentList,PyCallByClass
        return QCoreApplication.translate('CtrlCraft', message)

    def add_action(
            self,
            icon_path,
            text,
            callback,
            no_callback=False,
            params=None,
            enabled_flag=True,
            add_to_menu=False,
            add_to_toolbar=True,
            status_tip=None,
            whats_this=None,
            parent=None,
            checkable=False,
            toolbar=None):
        """Add a toolbar icon to the toolbar.

        :param icon_path: Path to the icon for this action. Can be a resource
            path (e.g. ':/plugins/foo/bar.png') or a normal file system path.
        :type icon_path: str

        :param text: Text that should be shown in menu items for this action.
        :type text: str

        :param callback: Function to be called when the action is triggered.
        :type callback: function

        :param enabled_flag: A flag indicating if the action should be enabled
            by default. Defaults to True.
        :type enabled_flag: bool

        :param add_to_menu: Flag indicating whether the action should also
            be added to the menu. Defaults to True.
        :type add_to_menu: bool

        :param add_to_toolbar: Flag indicating whether the action should also
            be added to the toolbar. Defaults to True.
        :type add_to_toolbar: bool

        :param status_tip: Optional text to show in a popup when mouse pointer
            hovers over the action.
        :type status_tip: str

        :param parent: Parent widget for the new action. Defaults None.
        :type parent: QWidgetfrom PyQt5 import QtGui, QtCore

        :param whats_this: Optional text to show in the status bar when the
            mouse pointer hovers over the action.

        :returns: The action that was created. Note that the action is also
            added to self.actions list.
        :rtype: QAction
        """

        # Create the dialog (after translation) and keep reference
        pix = QPixmap(os.path.join(
            os.path.dirname(__file__), 'img', icon_path))
        icon = QIcon(pix)
        action = QAction(icon, text, parent)

        if callback:
            if params:
                action.triggered.connect(partial(callback, params))
            else:
                action.triggered.connect(callback)

        action.setEnabled(enabled_flag)
        action.setCheckable(checkable)

        action.setData(params)

        if status_tip is not None:
            action.setStatusTip(status_tip)

        if whats_this is not None:
            action.setWhatsThis(whats_this)

        if add_to_toolbar:
            if not toolbar:
                toolbar = self.toolbar
            toolbar.addAction(action)

        if add_to_menu:
            self.iface.addPluginToMenu(
                self.name,
                action)

        self.actions.append(action)

        return action

    def addMenuEntry(self, name, callback, enabled=True, help_str=''):
        act = self.menu.addAction(name)
        if callback is not None:
            act.triggered.connect(callback)
            act.setEnabled(enabled)
            act.setToolTip(help_str)
        else:
            act.setEnabled(False)
            act.setToolTip("NOT INMPLEMENTED "+help_str)
        return act

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        self.initToolBarMobile()
        self.initMenu()

        if not self.isMobileVersion():
            self.initToolBarOffice()

        # Not ready by default. We're waiting to load a valid project
        self.setReady(False)

    def initToolBarMobile(self):
        if self.toolbar is None:
            self.toolbar = self.iface.addToolBar(self.name)
            self.toolbar.setObjectName('{} Office'.format(self.name))
        # Deposit selector
        self.depositCombo = QComboBox()
        self.depositCombo.setIconSize(
            QSize(int(self.iconSize), int(self.iconSize)))
        self.depositCombo.setSizePolicy(
            QSizePolicy.Preferred, QSizePolicy.Expanding)
        self.depositCombo.setSizeAdjustPolicy(QComboBox.AdjustToContents)

        self.depositComboAction = self.toolbar.addWidget(self.depositCombo)
        self.depositCombo.setToolTip(self.tr('Deposits'))

        self.depositCombo.currentIndexChanged.connect(self.depositSelected)

        # Panels selector
        self.panelCombo = QComboBox()
        self.panelCombo.setIconSize(
            QSize(int(self.iconSize), int(self.iconSize)))
        self.panelCombo.setSizePolicy(
            QSizePolicy.Preferred, QSizePolicy.Expanding)
        self.panelCombo.setSizeAdjustPolicy(QComboBox.AdjustToContents)

        self.panelComboAction = self.toolbar.addWidget(self.panelCombo)
        self.panelCombo.setToolTip(self.tr('Panels'))

        self.panelCombo.currentIndexChanged.connect(self.panelSelected)

        # Blast selector
        self.blastCombo = QComboBox()
        self.blastCombo.setIconSize(
            QSize(int(self.iconSize), int(self.iconSize)))
        self.blastCombo.setSizePolicy(
            QSizePolicy.Preferred, QSizePolicy.Expanding)
        self.blastCombo.setSizeAdjustPolicy(QComboBox.AdjustToContents)

        self.blastComboAction = self.toolbar.addWidget(self.blastCombo)
        self.blastCombo.setToolTip(self.tr('Blasts'))

        self.blastCombo.currentIndexChanged.connect(self.blastSelected)

        # Add some actions
        icon_path = 'icon3d.png'
        self.viewer_action = self.add_action(
            icon_path,
            text=self.tr('Display 2D/3D views'),
            callback=self.switch2D3D)

        icon_path = 'iconUndoSlab.png'
        self.undo_action = self.add_action(
            icon_path,
            text=self.tr('Undo'),
            callback=CtrlCraftTruck.unloadLastSlab,
            params=self)

        affect = CtrlCraftAffectRadiometrySlab(self)
        icon_path = 'iconAffectRadioSlab.png'
        self.affect_radio = self.add_action(
            icon_path,
            text=self.tr('Affect the grade on a slab'),
            callback=affect.activateMapTool,
            params=self)

        # Lithology
        self.lastRadiometrySelection = ''
        self.lithologies = {}

        # 3D viewer widget
        self.viewer3d = CtrlCraftViewer3D()
        self.viewer3d_dock = QDockWidget()
        self.viewer3d_dock.setTitleBarWidget(QWidget())
        self.viewer3d_dock.setWidget(self.viewer3d)

        # Add lithology buttons
        self.buildLithoButtons()

        # Trucks objects
        self.trucks = {}
        self.buildTruckButtons()
        # Build Select Truck Action
        icon_path = 'icon_select_truck.png'
        truckSelectAction = self.add_action(
            icon_path,
            text=self.tr('Select a truck'),
            callback=self.selectCurrentTruck)
        truckSelectAction.setCheckable(True)
        # Select the first truck by default
        if len(self.trucks) > 0:
            self.truckSelected(self.data.trucks_no_order()[0])

        # Radiometry dock widget
        self.radiometry_widget = CtrlCraftProbeRadiometryDock(self.iface)
        self.main_parameters_widget = CtrlCraftMainParametersDock(
            self.iface, self.data, self)
        self.main_parameters_widget.configModified.connect(
            self.updateGammaCorrection)

        # RTKube
        if not self.rtkube:
            self.rtkube = RTKubeProxy(self.iface, self.data,
                                      self.radiometry_widget,
                                      self.isMobileVersion())
            self.rtkube.init()

    def initMenu(self):
        self.menu = QMenu(self.name)

        self.action_open = self.addMenuEntry(
            self.tr('Open Project'), self.openProject)
        self.action_reset = self.addMenuEntry(
            self.tr('Reset Project'), self.resetProject)
        self.action_clear_db = self.addMenuEntry(
            self.tr('Clear Database'), self.clearDatabase)
        self.action_clear_db = self.addMenuEntry(
            self.tr('Define slab threshold symbology'), self.define_radiometry_threshold)

        self.menu.addSeparator()

        self.addMenuEntry(self.tr('Import Estimated Blocs'),
                          self.importEstimatedBlocs)
        self.addMenuEntry(self.tr('Import Slabs'), self.importSlabs)
        self.addMenuEntry(self.tr('Import Gate Data'), self.importGateData)
        self.addMenuEntry(self.tr('Import CO3'), self.importCO3)

        self.menu.addSeparator()
        self.addMenuEntry(self.tr('Filter levels'), self.filterLevels)
        self.addMenuEntry(self.tr('Remove levels'), self.removeLevel)
        self.addMenuEntry(self.tr('Revert removed levels'),
                          self.revertRemovedLevel)

        self.menu.addSeparator()
        self.addMenuEntry(self.tr('Import Database'), self.importDatabase)
        self.addMenuEntry(self.tr('Export Database'), self.exportDatabase)

        self.menu.addSeparator()

        self.addMenuEntry(self.tr('Export Trucks Loads Status'),
                          self.exportTrucksLoadsStatus)

        action = self.addMenuEntry(
            self.tr('Export RTKube data to CSV'), self.rtkube.export)
        self.rtkube.action_export = action

        self.iface.mainWindow().menuBar().addMenu(self.menu)

        self.menu.addSeparator()
        self.addMenuEntry(self.tr('Export Atlas PDF'), self.exportAtlasPDF)

        self.menu.addSeparator()
        self.addMenuEntry(self.tr('Exit QGIS'),
                          self.iface.actionExit().trigger)

    def initToolBarOffice(self):
        if self.toolbar_office == None:
            self.toolbar_office = self.iface.addToolBar(
                '{} Office'.format(self.name))
            self.toolbar_office.setObjectName('{} Office'.format(self.name))
        icon_path = 'iconOffice.png'
        self.add_action(
            icon_path,
            text=self.tr('CtrCraft Office Shipments'),
            callback=self.showOfficeShipmentsTable,
            toolbar=self.toolbar_office)
        icon_path = 'iconOfficeChart.png'
        self.add_action(
            icon_path,
            text=self.tr('CtrCraft Office Shipments Graph'),
            callback=self.showOfficeShipmentsChart,
            toolbar=self.toolbar_office)

        icon_path = 'iconOfficePanel.png'
        self.add_action(
            icon_path,
            text=self.tr('CtrCraft Office Estimated Blocs'),
            callback=self.showOfficeEstimatedBlocsTable,
            toolbar=self.toolbar_office)

        icon_path = 'iconOfficePanelChart.png'
        self.add_action(
            icon_path,
            text=self.tr('CtrCraft Office Estimated Blocs Graph'),
            callback=self.showOfficeEstimatedBlocsChart,
            toolbar=self.toolbar_office)

    def initQSettingsParameters(self):
        settings = QSettings()
        for i in range(1, 5):
            if settings.value('/CtrlCraft/Truck1/id') == None:
                settings.setValue(
                    '/CtrlCraft/Truck{}/id'.format(str(i)), str(i).zfill(3))
            if settings.value('/CtrlCraft/Truck1/sc') == None:
                settings.setValue('/CtrlCraft/Truck{}/sc'.format(str(i)), 8)
        if settings.value('/GeoCube/GPSmode') == None:
            settings.setValue('/GeoCube/GPSmode', 'Fixed')
        if settings.value('/GeoCube/angle') != None:
            settings.setValue('/GeoCube/angle', 90.0)
        if settings.value('/GeoCube/gammalow') == None:
            settings.setValue('/GeoCube/gammalow', 0.001)
        if settings.value('/GeoCube/gammaup') != None:
            settings.setValue('/GeoCube/gammaup', 0.001)

    def setReady(self, ready):
        for action in self.actions:
            action.setEnabled(ready)

        self.depositCombo.setEnabled(ready)
        self.panelCombo.setEnabled(ready)
        self.blastCombo.setEnabled(ready)
        self.depositSelected()
        self.blastSelected()

        if self.rtkube.ready:
            self.rtkube.action_export.setEnabled(ready)

        for truck in list(self.trucks.values()):
            truck.updateIconLoad()
            truck.updateIconUnload()

        if ready:
            self.fillDepositCombo()
            self.fillPanelCombo()
            self.fillBlastCombo()

            self.rtkube.layer = self.project.probeLayer()
            self.rtkube.topoLayer = self.project.topoLayer()

            if self.action_open:
                self.action_open.setEnabled(False)
                self.action_reset.setEnabled(True)
                self.action_clear_db.setEnabled(True)
        else:
            self.rtkube.layer = None
            self.depositCombo.clear()
            self.panelCombo.clear()
            self.blastCombo.clear()

            if self.action_open:
                self.action_open.setEnabled(True)
                self.action_reset.setEnabled(False)
                self.action_clear_db.setEnabled(False)

    def importEstimatedBlocs(self):
        """
        Imports Estimated Blocs from a CSV file
        """

        dlg = CtrlCraftOfficeEstimatedBlocsImport(
            self.data, self.projIniFile, self.currentDeposit)
        dlg.show()
        dlg.exec_()

        if dlg.updated:
            layer = self.project.estimatedBlocsLayer()
            layer.dataProvider().reloadData()
            self.iface.mapCanvas().refreshAllLayers()
            self.iface.mapCanvas().setExtent(layer.extent())

    def importSlabs(self):
        """
        Imports Slabs from a CSV file with SMU definition
        """

        dlg = CtrlCraftOfficeSlabsImport(
            self.data, self.projIniFile, self.currentDeposit)
        dlg.show()
        dlg.exec_()

        if dlg.updated:
            layer = self.project.slabsLayer()
            if layer is not None:
                layer.dataProvider().reloadData()
                self.iface.mapCanvas().refreshAllLayers()
                self.iface.mapCanvas().setExtent(layer.extent())

            self.fillDepositCombo()
            self.fillPanelCombo()
            self.fillBlastCombo()

    def importGateData(self):
        """
        Imports data from the gate.
        """

        dlg = CtrlCraftOfficeGateImport(self.data)
        dlg.show()
        dlg.exec_()

    def importCO3(self):
        """
        Import CO3 data.
        """

        dlg = CtrlCraftOfficeCO3Import(
            self.data, self.projIniFile, self.currentDeposit)
        dlg.show()
        dlg.exec_()

        if dlg.updated:
            layer = self.project.co3Layer()
            layer.dataProvider().reloadData()
            self.iface.mapCanvas().refreshAllLayers()
            self.iface.mapCanvas().setExtent(layer.extent())

    def resetProject(self):
        """
        Resets the current project to its initial stage.
        """

        self.project.reset()
        if self.project.exists():
            self.project.load()

    def openProject(self):
        """
        Opens the ctrlcraft project of not yet loaded
        """

        if not self.project.is_loaded():
            self.project.load()

    def clearDatabase(self):
        title = self.tr('Clear Database')
        msg = (self.tr('Are you sure you want to delete the database? Underlying data will be lost and RTKube connection will be closed.'))
        reply = QMessageBox.question(
            None, title, msg, QMessageBox.Yes | QMessageBox.No)

        if reply == QMessageBox.Yes:
            self.rtkube.layer = None
            self.rtkube.save = False
            self.project.close()

            self.updateStatus(self.tr('Clearing database...'))
            self.data.clear()
            self.updateStatus()

            self.project.load()
            self.iface.mapCanvas().refreshAllLayers()

            self.rtkube.layer = self.project.probeLayer()
            self.rtkube.topoLayer = self.project.topoLayer()
            self.rtkube.save = True

            if SIMU:
                CtrlCraftSimulator().gate_clear()

    def exportDatabase(self):

        dlg = CtrlCraftDatabaseExport(
            self.data, self.projIniFile, self.currentDeposit)
        dlg.show()
        dlg.exec_()

    def importDatabase(self):

        dlg = CtrlCraftDatabaseImport(
            self.data, self.rtkube, self.project, self.projIniFile, self.currentDeposit)
        dlg.show()
        dlg.exec_()

        if dlg.updated:
            self.iface.mapCanvas().refreshAllLayers()

            if SIMU:
                CtrlCraftSimulator().gate_clear()

    def exportAtlasPDF(self):

        dlg = ExportAtlas()
        dlg.show()
        dlg.exec_()

    def exportTrucksLoadsStatus(self):
        action = CtrlCraftTrucksLoadsStatusExport(self.data)
        action.export()

    def projectClosed(self):
        todo = True
        if CtrlCraft.isOnWindows():
            todo = self.name in active_plugins

        if todo:
            self.setReady(False)

    def projectLoaded(self):
        """
        Slot called when a new project has been loaded from desktop.
        """

        if CtrlCraft.isOnWindows() and self.name not in active_plugins:
            return

        if not self.project.is_loaded():
            self.setReady(False)
            self.data.close()
            return

        self.data.open()
        if not self.data.is_valid():
            return

        self.setReady(True)

    def filterLevels(self):
        layer = self.project.slabsLayer()
        if not layer:
            return

        originSubstring = layer.subsetString()
        idx = originSubstring.find(' and "centroid_z"')
        actualLevel = -9999
        if idx == -1:
            originSubstring += " "
            idx = len(originSubstring)
        else:
            lvlIdx = originSubstring.find(' <=')
            actualLevel = float(originSubstring[lvlIdx+4:])

        dlg = CtrlCraftLevelFilter(self.data,
                                   actualLevel,
                                   self.currentDeposit,
                                   self.currentPanel,
                                   self.currentBlast
                                  )
        dlg.show()

        res = dlg.exec_()
        if res:
            maxZ = dlg.mLevels.currentData()
            layer.setSubsetString(
                originSubstring[:idx] + ' and "centroid_z" <= ' + str(maxZ))

    def removeLevel(self):
        """
        Remove an entire level
        """
        layer = self.project.slabsLayer()
        if not layer:
            return

        dlg = CtrlCraftLevelFilter(self.data,
                                   0,
                                   self.currentDeposit,
                                   self.currentPanel,
                                   self.currentBlast
                                  )
        dlg.setWindowTitle(self.tr("Remove levels"))
        dlg.label.setText(self.tr("Remove levels:"))
        dlg.show()

        res = dlg.exec_()
        if res:
            self.data.remove_level(dlg.mLevels.currentData(),
                                   self.currentDeposit,
                                   self.currentPanel,
                                   self.currentBlast
                                  )
            self.iface.mapCanvas().refreshAllLayers()

    def revertRemovedLevel(self):
        """
        Revert a removed level
        """
        layer = self.project.slabsLayer()
        if not layer:
            return

        dlg = CtrlCraftLevelFilter(self.data,
                                   -1,
                                   self.currentDeposit,
                                   self.currentPanel,
                                   self.currentBlast
                                  )
        dlg.setWindowTitle(self.tr("Revert removed levels"))
        dlg.label.setText(self.tr("Revert removed levels:"))
        dlg.show()

        res = dlg.exec_()
        if res:
            self.data.revert_removed_level(dlg.mLevels.currentData(),
                                           self.currentDeposit,
                                           self.currentPanel,
                                           self.currentBlast
                                          )
            self.iface.mapCanvas().refreshAllLayers()

        dlg = CtrlCraftLevelFilter(self.data,
                                   -1,
                                   self.currentDeposit,
                                   self.currentPanel,
                                   self.currentBlast
                                  )

    def canvasRefreshed(self):
        # Check if the legend for the slab layer has changed, if so, update the 3D view
        # This is called every second as we have a timer for the geometer's position
        radiometrySelection = ""

        layer = self.project.slabsLayer()
        if not layer:
            return

        rend = layer.renderer()
        if rend.type() != "RuleRenderer":
            return

        style_path = os.path.join(os.path.dirname(__file__),'qgis/slabs_defaut.qml')
        _, rule_dict = CtrlCraftSlabThreshold.get_threshold_values(style_path)
        p = re.compile(r'\d+[\.?]\d+')
        labels = list([a for a in rule_dict.keys() if rule_dict[a][3] != 1])

        for rule in rend.legendSymbolItems():
            # The rule MUST be fixed as follow
            if rule.label() in labels:
                if rule.label().startswith("+"):
                    minv = float(rule.label())
                    radiometrySelection += " and (radiometry < {}) ".format(str(minv))
                else:
                    minv, maxv = [float(v) for v in rule.label().split(' - ')]
                    radiometrySelection += " and (radiometry < {} or radiometry > {}) ".format(str(minv), str(maxv))

        if self.lastRadiometrySelection != radiometrySelection:
            self.run3d(radiometrySelection)

        self.lastRadiometrySelection = radiometrySelection

    def fillDepositCombo(self):
        """ Fill the deposit combo """
        self.depositCombo.clear()
        for r in self.data.deposits():
            self.depositCombo.addItem(r, r)

    def fillPanelCombo(self):
        """ Fill the panel combo """
        self.panelCombo.clear()
        for r in self.data.panels(self.currentDeposit):
            self.panelCombo.addItem(r, r)

    def fillBlastCombo(self):
        """ Fill the blast combo """
        self.blastCombo.clear()
        for r in self.data.blasts(self.currentPanel):
            self.blastCombo.addItem(r, r)

    def buildTruckButtons(self):
        """ Build the truck buttons """
        for truckid in self.data.trucks_no_order():
            # Create a new truck
            truck = CtrlCraftTruck(truckid, self)
            self.trucks[truckid] = truck

        self.truckUpdated()

    def selectCurrentTruck(self):
        """
        Open Dialog to select a Truck
        """
        dlg = QDialog()
        vLayout = QHBoxLayout(dlg)
        for truck in self.trucks.values():
            TPushButton = QPushButton(truck.id)
            TPushButton.clicked.connect(
                lambda state, x=truck.id: self.truckSelected(x))
            TPushButton.clicked.connect(dlg.close)
            vLayout.addWidget(TPushButton)

        res = dlg.exec_()

    def truckSelected(self, id):
        """ Change truck, triggered using truck combobox """
        if self.truck:

            self.truck.updated.disconnect(self.truckUpdated)
            for i in self.actions:
                if i.text() in [self.tr('Load the truck ')+'{}'.format(self.truck.id),
                                self.tr('Unload the truck ') +
                                '{}'.format(self.truck.id),
                                self.tr('Probe grade for truck ')+'{}'.format(self.truck.id)]:
                    self.toolbar.removeAction(i)

        truck = self.trucks[id]

        truckAction = self.add_action(
            truck.loadIcon(),
            text=self.tr('Load the truck ')+'{}'.format(truck.id),
            callback=truck.activateMapTool)
        truckAction.setCheckable(True)
        truck.actionLoad = truckAction

        truckEmptyAction = self.add_action(
            truck.unloadIcon(),
            text=self.tr('Unload the truck ')+'{}'.format(truck.id),
            callback=truck.removeLoads)
        truck.actionUnload = truckEmptyAction

        truckStateAction = self.add_action(
            truck.stateIcon(),
            text=self.tr('Probe grade for truck ')+'{}'.format(truck.id),
            callback=None
        )
        truck.actionState = truckStateAction

        # Store truck
        self.truck = truck
        self.truck.updated.connect(self.truckUpdated)

        # Re-activate load maptool
        if type(self.iface.mapCanvas().mapTool()) == CtrlCraftTruckLoadMapTool:
            self.truck.activateMapTool()

    def buildLithoButtons(self):
        """ Build the litho list and buttons to apply to slab """
        # List
        for name in self.data.lithologies():
            lithology = CtrlCraftLithology(self, name)
            # Action
            icon_path = 'iconLithoSlab_{}.png'.format(text_to_id(name))
            action = self.add_action(
                icon_path,
                text=self.tr('Set lithology ') + '{}'.format(name),
                callback=lithology.activateMapTool)
            action.setCheckable(True)
            lithology.action = action

            self.lithologies[name] = lithology

    def depositSelected(self):
        CtrlCraftLogger().log()

        self.fillPanelCombo()
        self.fillBlastCombo()

        if self.currentDeposit:
            self.project.setDepositGroupName(self.currentDeposit)
            self.updateGammaCorrection()
        else:
            self.project.setDepositGroupName('None')

    def updateGammaCorrection(self):
        settings = QSettings()
        gamma_config = configparser.ConfigParser()
        gamma_config.read(self.gammaIniFile)
        try:
            settings.setValue('/GeoCube/gammaup',
                              float(gamma_config[self.currentDeposit]['teneur_sonde_haute'].split('*')[0]))
            settings.setValue('/GeoCube/gammalow',
                              float(gamma_config[self.currentDeposit]['teneur_sonde_basse'].split('*')[0]))
        except KeyError:
            self.logger.log(
                'No custom gamma correction known, using standard value')
            settings.setValue('/GeoCube/gammaup', 0.001)
            settings.setValue('/GeoCube/gammalow', 0.001)

    def panelSelected(self):
        # Update blast combo
        self.fillBlastCombo()

    def blastSelected(self):
        layer = self.project.slabsLayer()
        if layer:
            if self.currentBlast:
                layer.setName(
                    "{panel}-{blast}".format(panel=self.currentPanel, blast=self.currentBlast))
                self.run3d()
                self.filterLayer()
            else:
                layer.setName("None".format(
                    panel=self.currentPanel, blast=self.currentBlast))
        levelsLayer = self.project.levelsLayer()
        if levelsLayer:
            if self.currentBlast:
                self.filterLevelByBlast()


    def filterLayer(self):
        """ Filter the slabs layer according to current panel and blast"""
        layer = self.project.slabsLayer()

        if not layer:
            return

        layer.setSubsetString("deposit='{deposit}' and panel='{panel}' and blast='{blast}'"
                              .format(panel=self.currentPanel,
                                      blast=self.currentBlast,
                                      deposit=self.currentDeposit))
        layer.triggerRepaint()
        extent = layer.extent()
        self.canvas.setExtent(extent)
        self.canvas.zoomOut()

    def filterLevelByBlast(self):
        layer = self.project.levelsLayer()

        if not layer:
            return

        layer.setSubsetString("deposit='{deposit}' and panel='{panel}' and blast='{blast}'"
                              .format(panel=self.currentPanel,
                                      blast=self.currentBlast,
                                      deposit=self.currentDeposit))

    def run3d(self, radiometrySelection=''):
        """ Run 3D view """

        # Reset the 3D view
        conn_string = None
        if self.data.is_valid() \
                and self.project.slabsLayer().featureCount() > 0:
            conn_string = self.data.conn_string

        self.viewer3d.resetScene(conn_string,
                                 panel=self.currentPanel,
                                 blast=self.currentBlast,
                                 radSel=radiometrySelection)

    def switch2D3D(self, params):
        """ Switch beetween 2D and 3D view """
        mainWindow = self.iface.mainWindow()
        centralWidget = mainWindow.findChild(QWidget, 'centralwidget')

        if not self.viewSize:
            self.viewSize = centralWidget.size()
            self.viewMinSize = centralWidget.minimumSize()
            self.viewMaxSize = centralWidget.maximumSize()

        if self.viewMode == 0:
            # Currently 2D and 3D, we want to show 3D only
            self.iface.removeDockWidget(self.viewer3d_dock)

            newMaxSize = QSize(self.viewMaxSize.width(), 0)
            centralWidget.setMaximumSize(newMaxSize)

            currentMinWidth = self.viewer3d_dock.minimumSize().width()
            newMinSize = QSize(currentMinWidth, self.viewSize.height())
            self.viewer3d_dock.setMinimumSize(newMinSize)
            self.iface.addDockWidget(Qt.TopDockWidgetArea, self.viewer3d_dock)

            self.viewMode = 1
        elif self.viewMode == 1:
            # Currently only 3D, and we want show only 2D
            self.iface.removeDockWidget(self.viewer3d_dock)

            centralWidget.setMinimumSize(self.viewMinSize)
            centralWidget.setMaximumSize(self.viewMaxSize)

            self.viewMode = 2
        else:
            # Currently only 2D and we want to show 2D and 3D
            self.iface.removeDockWidget(self.viewer3d_dock)

            self.viewer3d_dock.setMinimumSize(self.viewMinSize)
            self.iface.addDockWidget(
                Qt.BottomDockWidgetArea, self.viewer3d_dock)

            self.viewMode = 0

    def unload(self):
        """ Removes the plugin menu item and icon from QGIS GUI. """
        for action in self.actions:
            self.iface.removePluginMenu(
                self.name,
                action)
            self.iface.removeToolBarIcon(action)

        # remove the toolbar
        self.toolbar and self.toolbar.setParent(None)
        self.viewer3d_dock.setParent(None)

        if self.data:
            self.data.close()

        if self.timer:
            self.timer.stop()

        if self.viewer3d:
            self.viewer3d.close()

    def showOfficeShipmentsTable(self, params):
        """ Office management """
        dlg = CtrlCraftOfficeShipmentsTable(self.data)
        dlg.init(self.currentDeposit, self.currentPanel, self.currentBlast)
        dlg.show()
        dlg.exec_()

    def showOfficeEstimatedBlocsTable(self, params):
        """ Office management """
        dlg = CtrlCraftOfficeEstimatedBlocsTable(self.data)
        dlg.init(self.currentDeposit, self.currentPanel, self.currentBlast)
        dlg.show()
        dlg.exec_()

    def showOfficeShipmentsChart(self, params):
        """ Office chart management """
        max_slab = 0
        for truck in self.trucks.values():
            max_slab = max(max_slab, truck.max_slab)
        max_slab = max(max_slab, self.data.get_max_slabs_shipment(
            self.currentDeposit, self.currentPanel, self.currentBlast))
        dlg = CtrlCraftOfficeShipmentsChart(self.data, self.canvas, max_slab)
        dlg.init(self.currentDeposit, self.currentPanel, self.currentBlast)
        dlg.show()
        dlg.exec_()

    def showOfficeEstimatedBlocsChart(self, params):
        """ Office chart management """
        dlg = CtrlCraftOfficeEstimatedBlocsChart(self.data, self.canvas)
        dlg.init(self.currentDeposit, self.currentPanel, self.currentBlast)
        dlg.show()
        dlg.exec_()

    def updateStatus(self, msg=''):
        bar = self.iface.mainWindow().statusBar()

        if msg:
            self.status_label = QLabel()
            self.status_bar = QProgressBar()
            self.status_bar.setMinimum(0)
            self.status_bar.setMaximum(0)

            self.status_label.setText(msg)
            bar.insertWidget(0, self.status_label)
            bar.insertWidget(1, self.status_bar)
        else:
            bar.removeWidget(self.status_label)
            bar.removeWidget(self.status_bar)

        qApp.processEvents()

    def define_radiometry_threshold(self):
        if not self.project.is_loaded():
            return

        dlg = CtrlCraftSlabThreshold(self.project, self.data)
        dlg.accepted.connect(self.refresh)
        dlg.exec_()

    @staticmethod
    def isMobileVersion():
        v = QSettings().value('/CtrlCraft/mobile', '')
        if v == 'true' or v == 'True':
            return True
        else:
            return False

    @staticmethod
    def isOnWindows():
        return platform.system() == 'Windows'
