import os

import serial.tools.list_ports

from PyQt5 import QtWidgets, uic
from PyQt5.QtWidgets import QHBoxLayout, QLabel, QLineEdit, QSpinBox
from PyQt5.QtCore import Qt, QSettings, pyqtSignal
from PyQt5.QtGui import QIntValidator

from .custom_edition import CtrlCraftCustomEdition

FORM_CLASS, _ = uic.loadUiType(os.path.join(
    os.path.dirname(__file__), 'ui', 'main_parameters.ui'))

settings = QSettings()


class CtrlCraftMainParametersDock(QtWidgets.QDockWidget, FORM_CLASS):

    configModified = pyqtSignal()

    def __init__(self, iface, data, app, parent=None):
        """Constructor."""
        super(CtrlCraftMainParametersDock, self).__init__(parent)
        self.iface = iface
        self.data = data
        self.app = app
        self.setupUi(self)
        self.trucks_info = {}

        if 'CTRLCRAFT_ADVANCEDUSER' in os.environ:
            if int(os.environ['CTRLCRAFT_ADVANCEDUSER']) == True:
                self.mProjButton.setEnabled(True)
                self.mGammaButton.setEnabled(True)
        else :
            self.mProjButton.setEnabled(False)
            self.mGammaButton.setEnabled(False)
        self.mCancelButton.clicked.connect(self.__close)
        self.mProjButton.clicked.connect(self.__editProj)
        self.mGammaButton.clicked.connect(self.__editGamma)
        self.mSaveButton.clicked.connect(self.__save)

        self.nbTruckSpinBox.valueChanged.connect(self.__trucksUpdate)
        self.nbTruckSpinBox.setValue(int(settings.value('/CtrlCraft/nbTrucks', 4)))
        self.__update()
        self.iface.addDockWidget(Qt.LeftDockWidgetArea, self)
        self.isFloating()
        self.hide()


    def __close(self):
        self.close()

    def __populateGPSComboBox(self, current_value):
        gps_dict={0: 'Undetermined', 1: 'GNSS', 6: 'Diff', 7: 'DGNSS', 15: 'Float', 23: 'Fixed'}
        for mode in gps_dict.values():
            self.mGPSComboBox.addItem(mode)
        self.mGPSComboBox.setCurrentIndex(self.mGPSComboBox.findText(current_value) if self.mGPSComboBox.findText(current_value) != -1 else 0)

    def __populateSerialComboBoxes(self):
        self.mSerialPortRover.clear()
        self.mSerialPortBase.clear()
        self.mSerialPortSpare.clear()
        comports = serial.tools.list_ports.comports()
        self.mSerialPortRover.addItem("")
        self.mSerialPortBase.addItem("")
        self.mSerialPortSpare.addItem("")
        for i, com in enumerate(comports):
            self.mSerialPortRover.addItem(com[0])
            self.mSerialPortBase.addItem(com[0])
            self.mSerialPortSpare.addItem(com[0])
            information = "Ports: {}\nDescription: {}\nHWID: {}".format(*com)
            self.mSerialPortRover.setItemData(i+1, information, Qt.ToolTipRole)
            self.mSerialPortBase.setItemData(i+1, information, Qt.ToolTipRole)
            self.mSerialPortSpare.setItemData(i+1, information, Qt.ToolTipRole)

        self.mSerialPortRover.setCurrentIndex(self.mSerialPortRover.findText(str(settings.value('/GeoCube/serialportrover'))) if self.mSerialPortRover.findText(str(settings.value('/GeoCube/serialportrover'))) != -1 else 0)
        self.mSerialPortBase.setCurrentIndex(self.mSerialPortBase.findText(str(settings.value('/GeoCube/serialportspare'))) if self.mSerialPortBase.findText(str(settings.value('/GeoCube/serialportbase'))) != -1 else 0)
        self.mSerialPortSpare.setCurrentIndex(self.mSerialPortSpare.findText(str(settings.value('/GeoCube/serialportspare'))) if self.mSerialPortSpare.findText(str(settings.value('/GeoCube/serialportspare'))) != -1 else 0)

    def __editGamma(self):
        """ Open a edit windows for editing proj.ini file """
        gamma_path = os.path.join(os.path.dirname(__file__),'..','gamma.ini')
        edit = CtrlCraftCustomEdition(gamma_path)
        self.mProjButton.setEnabled(False)
        self.mSaveButton.setEnabled(False)
        self.mCancelButton.setEnabled(False)
        edit.show()
        edit.exec_()
        self.configModified.emit()
        self.mProjButton.setEnabled(True)
        self.mSaveButton.setEnabled(True)
        self.mCancelButton.setEnabled(True)

    def __editProj(self):
        """ Open a edit windows for editing proj.ini file """
        proj_path = os.path.join(os.path.dirname(__file__),'..','proj.ini')
        edit = CtrlCraftCustomEdition(proj_path)
        self.mProjButton.setEnabled(False)
        self.mSaveButton.setEnabled(False)
        self.mCancelButton.setEnabled(False)
        edit.show()
        edit.exec_()
        self.configModified.emit()
        self.mProjButton.setEnabled(True)
        self.mSaveButton.setEnabled(True)
        self.mCancelButton.setEnabled(True)

    def __validate(self):
        validator = QIntValidator(000,999)
        res=[]
        for i in range(1, self.nbTruck+1) :
            res.append([validator.validate(self.trucks_info["Truck {}".format(i)][0].text(),2)[0],len(self.trucks_info["Truck {}".format(i)][0].text())])
        test = True
        for i in res:
            if i != [2,3]:
                test = False
        if test == False :
            QtWidgets.QMessageBox.warning(None, self.tr("Truck ID"),
                self.tr("Truck ID must be a integer of 3 digits"))
        return test

    def __cleanHBoxes(self):
        for layout in self.TVerticalLayout.children() :
            for index in reversed(range(layout.count())) :
                layout.itemAt(index).widget().setParent(None)

            layout = self.TVerticalLayout.itemAt(0)
            self.TVerticalLayout.removeItem(layout)
            layout.setParent(None)
            # del layout


    def __trucksUpdate(self):
        self.nbTruck = self.nbTruckSpinBox.value()
        self.trucks_info={}
        self.__cleanHBoxes()
        for i in range(1, self.nbTruck+1):
            hBox = QHBoxLayout()
            hBox.addWidget(QLabel(self.tr('Truck ')+'{}'.format(i)))
            lineEdit = QLineEdit()
            lineEdit.setValidator(QIntValidator(000,999))
            lineEdit.setText(settings.value( '/CtrlCraft/Truck{}/id'.format(i), str(i).zfill(3)))
            hBox.addWidget(lineEdit)
            spinBox = QSpinBox()
            spinBox.setValue(int(settings.value( '/CtrlCraft/Truck{}/sc'.format(i), 8)))
            hBox.addWidget(spinBox)
            self.trucks_info["Truck {}".format(i)]=(lineEdit,spinBox)
            self.TVerticalLayout.addLayout(hBox)

    def __update(self):
        """ Init parameters in the different boxes """
        self.__trucksUpdate()
        self.__populateGPSComboBox(str(settings.value( '/GeoCube/GPSmode')))
        self.mAngleSpinBox.setValue(float(settings.value( '/GeoCube/angle', 10)))
        self.__populateSerialComboBoxes()

    def __save(self):
        """ Save new parameters and actualize database"""
        test = self.__validate()
        if test == True :
            settings.setValue('/CtrlCraft/nbTrucks',self.nbTruck)
            for i in range(1, self.nbTruck+1):
                settings.setValue( '/CtrlCraft/Truck{}/id'.format(i),self.trucks_info["Truck {}".format(i)][0].text())
                settings.setValue( '/CtrlCraft/Truck{}/sc'.format(i),self.trucks_info["Truck {}".format(i)][1].value())
            settings.setValue( '/GeoCube/GPSmode',self.mGPSComboBox.currentText())
            settings.setValue( '/GeoCube/angle',self.mAngleSpinBox.value())
            settings.setValue('/GeoCube/serialportrover', self.mSerialPortRover.currentText())
            settings.setValue('/GeoCube/serialportbase', self.mSerialPortBase.currentText())
            settings.setValue('/GeoCube/serialportspare', self.mSerialPortSpare.currentText())
            self.data.truck_ids_change([self.trucks_info["Truck {}".format(i)][0].text() for i in range(1, self.nbTruck+1)])
            self.configModified.emit()
            self.app.toolbar.close()
            self.app.toolbar_office.close()
            self.app.toolbar = None
            self.app.initToolBarMobile()
            self.app.toolbar_office = None
            self.app.initToolBarOffice()
            self.app.resetProject()
            self.close()
